char rcsid[] = "$Id: acctd.c,v 1.6 1996/10/15 16:28:00 bwm Exp $";
/*
 * $Log: acctd.c,v $
 * Revision 1.6  1996/10/15  16:28:00  bwm
 *          :)
 *
 * Revision 1.5  1996/10/14  00:01:35  bwm
 *  , -  acctd   core   
 * (   ).   *db
 *   .
 *
 * Revision 1.4  1996/10/09  15:27:23  bwm
 *     
 *
 * Revision 1.3  1996/10/09  13:14:13  bwm
 *    utmp  /var/adm/utmp
 *
 * Revision 1.2  1996/10/09  13:12:41  bwm
 *      nas   -.
 *
 * Revision 1.1  1996/10/09  13:02:45  bwm
 * Initial revision
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <syslog.h>
#include <sysexits.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/errno.h>
#if defined(__linux__)
#include <sys/file.h>
#endif
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#include <ndbm.h>

#include <utmp.h>
#include "own_wtmp.h"

#include "Config.h"


char *pname;
char *utmp_path = _PATH_UTMP;
char *dir_path = NULL;
int utmp_file = -1;
time_t scan_time = SCANTIME;
time_t reinit = REINIT;
char *killer = ACTPROG;
char *dbfile = DBFILE;

int debug = 0;

DBM *db = NULL;
datum key, val;

time_t told, tnow;

void usage()
{
    fprintf(stderr, "\
Usage: %s [-d] [-w file] [-r [dir]] [-t time] [-a [time]] [-k path]\n\
 -d          - run in debug mode (do not daemonize)\n\
 -w <file>   - use utmp file <file> instead of %s\n\
 -u <file>   - use user database file <file> instead of %s\n\
 -r [<dir>]  - place restrict files in <dir> (%s as default) directory\n\
 -t <time>   - scan wtmp each <time> seconds instead of %d\n\
 -a <time>   - kill users with great than <time> seconds on-line time\n\
               (if <time>==0, then never) instead of %d\n\
 -k <path>   - path to killer program instead of %s\n\
", pname, _PATH_UTMP, DBFILE, DENYDIR, SCANTIME, REINIT, ACTPROG);
    exit(EX_USAGE);
}

void hangup(s)
{
    syslog(LOG_ALERT, "got %d signal -- exiting", s);
    if (debug)
	printf("Got %d signal -- exiting\n", s);
    if (db)
	dbm_close(db);
    closelog();
    exit(EX_OK);
}

void killuser(u, l, h, act)
char *u, *l, *h;
{
    int pid;
    if ((pid = fork())) {
	int c;
	waitpid(pid, &c, 0);
	return;
    }
#if defined(__FreeBSD__) || defined(__bsdi__)
    daemon(1, 0);
#elif defined (__linux__)
    setsid();
#endif
    if ((pid = fork())) {
	sleep(1);
	exit(EX_OK);
    }
#if defined(__FreeBSD__) || defined(__bsdi__)
    daemon(1, 0);
#elif defined (__linux__)
    setsid();
#endif
    if (dir_path) {
	time_t t;
	char lck[256];
	FILE *f;

	time(&t);
	if (dir_path && act) {
	    if(sizeof(lck) <= strlen(dir_path) + strlen(u) + 1)
	     {
	      syslog(LOG_ERR, "can't create deny file for %s: internal buffer overflow", u);
	      if(debug)
  	        printf("Can't create deny file for %s: internal buffer overflow\n", u);
	     }
	    else
	     {
  	      sprintf(lck, "%s/%s", dir_path, u);
	      if ((f = fopen(lck, "w"))) {
	  	fprintf(f, "\n*** Sorry, your time limit is exceeded at %s", ctime(&t));
		fclose(f);
	      } else {
		syslog(LOG_ERR, "can't create deny file %s: %m", lck);
 	        if(debug)
		  printf("Can't create deny file %s: %s\n", lck, strerror(errno));
              }
             }
	}
    }
    execl(killer, "-killuser", u, l, h, NULL);
    syslog(LOG_ERR, "can't exec %s %s %s %s: %m", killer, u, l, h);
    if(debug)
      printf("Can't exec %s %s %s %s: %s\n", killer, u, l, h, strerror(errno));
    exit(EX_OSERR);
}

int db_update(dbp, t, u, l, h)
DBM *dbp;
char *u, *l, *h;
time_t t;
{
#ifdef DBF_1
    TQ_st tq;
    TQ_st *tqp;
#else
    time_t tq;
#endif

    key.dptr = u;
    key.dsize = strlen(u) + 1;
    val = dbm_fetch(dbp, key);
    if (!val.dptr)
	return -1;

#ifdef DBF_1	

    tqp = (TQ_st *) val.dptr;
    tq.tl  = tqp->tl;
    tq.tll = tqp->tll + t;
    tq.dl  = tqp->dl;
    tq.dll = tqp->dll + t;
    tq.tot = tqp->tot - t;
    if (tq.dll <= 0) {
      if (tq.dl >= 0) {
      	killuser(u, l, h, 1);
	syslog(LOG_INFO,
	       "user %s on line %s nas %s day time limit reached", u, l, h);
	if (debug)
	    printf("User %s on line %s nas %s day time limit reached\n", u, l, h);
	goto e1;
      }
    }
    if (tq.tll <= 0) {
      if (tq.tl >= 0) {
      	killuser(u, l, h, 1);
	syslog(LOG_INFO,
	       "user %s on line %s nas %s total time limit reached", u, l, h);
	if (debug)
	    printf("User %s on line %s nas %s total time limit reached\n", u, l, h);
      }
    }

e1:

#else

  tq = *(time_t *)val.dptr;
  tq += t;
  if (tq < 0) {
	killuser(u, l, h, 1);
  	syslog(LOG_INFO,
		"user %s on line %s nas %s time limit reached", u, l, h);
	if (debug)
	  printf("User %s on line %s nas %s time limit reached\n", u, l, h);
	tq = 0;
  }

#endif

    val.dptr = (void *) &tq;
    val.dsize = sizeof(tq);
    return dbm_store(db, key, val, DBM_REPLACE);
}


int main(ac, av)
char *av[];
{
    int ch;
    struct utmp w;
#ifdef __linux__
    int nf;
#endif

    pname = strrchr(av[0], '/');
    pname = pname ? ++pname : av[0];

    opterr = 0;
    while ((ch = getopt(ac, av, ":dw:r:t:a:k:")) != EOF) {
        if(ch == ':')
          ch = optopt;
	switch (ch) {
	case 'w':
	    if(!optarg)
	      usage();
	    utmp_path = optarg;
	    break;
	case 'u':
	    if(!optarg)
	      usage();
	    dbfile = optarg;
	    break;
	case 'r':
	    dir_path = optarg ? optarg : DENYDIR;
	    break;
	case 't':
	    if(!optarg)
	      usage();
	    scan_time = atoi(optarg);
	    break;
	case 'a':
	    if(!optarg)
	      usage();
	    reinit = atoi(optarg);
	    break;
	case 'k':
	    if(!optarg)
	      usage();
	    killer = optarg;
	    break;
	case 'd':
	    debug = 1;
	    break;
	default:
	    usage();
	    break;
	}
    }
    ac -= optind;
    av += optind;

    openlog(pname, LOG_PID | LOG_NDELAY, LOG_FACIL);

    utmp_file = open(utmp_path, O_RDONLY);
    if (utmp_file < 0) {
	fprintf(stderr, "Unable to open wtmp file %s: %s\n", utmp_path, strerror(errno));
	syslog(LOG_ALERT,
	       "unable to open wtmp file %s", utmp_path);
	exit(EX_OSFILE);
    }
    signal(SIGHUP, hangup);
    signal(SIGINT, hangup);
    signal(SIGQUIT, hangup);
    signal(SIGTERM, hangup);

    syslog(LOG_INFO, "started (version %s)", VERSION);
    if(debug)
      printf("Started (version %s)\n", VERSION);

    if (!debug) {
#if defined(__FreeBSD__) || defined(__bsdi__)
	daemon(1, 0);
#elif defined (__linux__)
	nf = open("/dev/null", O_RDWR, 0600);
	dup2(nf, 0);
	dup2(nf, 1);
	dup2(nf, 2);
	close(nf);
	setsid();
	if (fork())
	    exit(0);
#endif
    }
    told = tnow = time(NULL);
    while (1) {
	flock(utmp_file, LOCK_UN);
	sleep(scan_time);
	lseek(utmp_file, 0L, SEEK_SET);
	if (flock(utmp_file, LOCK_EX | LOCK_NB) < 0)
	    if (errno == EWOULDBLOCK) {
		syslog(LOG_WARNING,
		       "cannot lock utmp file");
		if (debug)
		    printf("Cannot lock utmp file\n");
		continue;
	    }
	db = dbm_open(dbfile, O_RDWR | O_EXLOCK, 0600);
	if (!db) {
	    fprintf(stderr, "Unable to open user database %s: %s\n", dbfile, strerror(errno));
	    syslog(LOG_ALERT,
		   "unable to open user database %s", dbfile);
	    exit(EX_OSFILE);
	}
	tnow = time(NULL);
	while (read(utmp_file, &w, sizeof(w)) == sizeof(w)) {
	    char utname[UT_NAMESIZE + 1];
	    char utline[UT_LINESIZE + 1];
	    char uthost[UT_HOSTSIZE + 1];
	    if (!*w.ut_name)
		continue;
	    strncpy(utname, w.ut_name, UT_NAMESIZE);
	    utname[UT_NAMESIZE] = '\0';
	    strncpy(utline, w.ut_line, UT_LINESIZE);
	    utname[UT_LINESIZE] = '\0';
	    strncpy(uthost, w.ut_host, UT_HOSTSIZE);
	    utname[UT_HOSTSIZE] = '\0';
	    if (db_update(db, told - tnow, utname, utline, uthost) < 0) {
/*
                  syslog(LOG_INFO, "no such user: %s", utname);
 */
                  if(debug)
                    printf("No such user: %s\n", utname);
	    }
/* Patch by dyer@sut.ru */
            if (reinit && (tnow - w.ut_time) > reinit) {
                 syslog(LOG_INFO,
                        "user %s on line %s nas %s logged in too long",
                        utname, utline, uthost);
		 if (debug)
                   printf("User %s on line %s nas %s logged in too long\n",
                          utname, utline, uthost);
                 killuser(utname, utline, uthost, 0);
            }
	}
	told = tnow;
	dbm_close(db);
	db = NULL;
    }
    exit(EX_OSERR);
}
